---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

---@class Bastion.TimeToDieUnit
local TimeToDieUnit = {
    Enums = {
        Duration = {
            DEFAULT_TTD = 30,
            FOREVER = 300,
            TRIVIAL = 5,
        },
        DeathPercent = {
            [162099] = 0.5, -- General Kaal; Sanguine Depths
            [166608] = 0.1, -- Mueh'zala; De Other Side
            [164929] = 0.2, -- Tirnenn Villager; Mists of Tirna Scythe
            [164804] = 0.2, -- Droman Oulfarran; Mists of Tirna Scythe
        },
        Excluded = {
            [23775]  = true,   -- Head of the Horseman
            [120651] = true,   -- Explosives
            [156227] = true,   -- Neferset Denizen
            [160966] = true,   -- Thing from Beyond?
            [161895] = true,   -- Thing from Beyond?
            [157452] = true,   -- Nightmare Antigen in Carapace
            [158041] = 310126, -- N'Zoth with Psychic Shell
            [164698] = true,   -- Tor'ghast Junk
            [177117] = 355790, -- Ner'zhul: Orb of Torment (Protected by Eternal Torment)
            [176581] = true,   -- Painsmith:  Spiked Ball
            [186150] = true,   -- Soul Fragment (Gavel of the First Arbiter)
            [185685] = true,   -- Season 3 Relics
            [185680] = true,   -- Season 3 Relics
            [185683] = true,   -- Season 3 Relics
            [183501] = 367573, -- Xy'mox: Genesis Bulwark
            [166969] = true,   -- Frieda
            [166970] = true,   -- Stavros
            [166971] = true,   -- Niklaus
            [168113] = 329606, -- Grashaal (when shielded)
            [168112] = 329636, -- Kaal (when shielded)
            [193760] = true,   -- Surging Ruiner (Raszageth) -- gives bad range information.
            [204560] = true,   -- Incorporeal Being
        }
    },
}

TimeToDieUnit.__index = TimeToDieUnit

---@param recycledUnit Bastion.TimeToDieUnit | false
---@param when number
---@param healthPct number
---@param unit? WowGameObject
function TimeToDieUnit:New(recycledUnit, when, healthPct, unit)
    ---@class Bastion.TimeToDieUnit
    local self = recycledUnit and recycledUnit:Recycle() or setmetatable({}, TimeToDieUnit)
    self.firstSeen = when
    self.firstHealthPct = healthPct

    self.lastSeen = when
    self.lastHealthPct = healthPct

    self.unit = unit or false

    self.rate = 0
    self.n = 0

    self.npcid = unit and unit:id() or false

    self.deathPercent = self.npcid and TimeToDieUnit.Enums.DeathPercent[self.npcid] or 0
    self.deathTime = self.unit and (UnitIsTrivial(self.unit:unit()) and UnitLevel(self.unit:unit()) > -1) and
        TimeToDieUnit.Enums.TRIVIAL or TimeToDieUnit.Enums.DEFAULT_TTD

    self.excluded = self.npcid and TimeToDieUnit.Enums.Excluded[self.npcid] or false
    return self:Update(when, healthPct)
end

function TimeToDieUnit:Recycle()
    self.firstSeen = 0
    self.firstHealthPct = 0
    self.lastSeen = 0
    self.lastHealthPct = 0
    self.unit = false
    self.rate = 0
    self.n = 0
    self.npcid = 0
    self.deathPercent = 0
    self.deathTime = 0
    self.excluded = false
    return self
end

---@param when number
---@param healthPct number
function TimeToDieUnit:Update(when, healthPct)
    local difference = self.lastHealthPct - healthPct

    -- We don't recalculate the rate when enemies heal.
    if difference > 0 then
        local elapsed = when - self.lastSeen

        -- If this is our first health difference, just store it.
        if self.n == 0 then
            self.rate = difference / elapsed
            self.n = 1
        else
            local samples = min(self.n, 9)
            local newRate = self.rate * samples + (difference / elapsed)
            self.n = samples + 1
            self.rate = newRate / self.n
        end

        self.deathTime = (healthPct - self.deathPercent) / self.rate
    end

    self.lastHealth = healthPct
    self.lastSeen = when
    return self
end

Bastion.TimeToDieUnit = TimeToDieUnit
